"use strict";
/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Sebastian Bittrich <sebastian.bittrich@rcsb.org>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.isHydrophobic = exports.isInMembranePlane = exports.calculate = exports.computeANVIL = exports.ANVILParams = void 0;
var tslib_1 = require("tslib");
var structure_1 = require("../../mol-model/structure");
var mol_task_1 = require("../../mol-task");
var centroid_helper_1 = require("../../mol-math/geometry/centroid-helper");
var accessible_surface_area_1 = require("../../mol-model-props/computed/accessible-surface-area");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var util_1 = require("../../mol-model/structure/util");
var shrake_rupley_1 = require("../../mol-model-props/computed/accessible-surface-area/shrake-rupley");
var param_definition_1 = require("../../mol-util/param-definition");
;
exports.ANVILParams = {
    numberOfSpherePoints: param_definition_1.ParamDefinition.Numeric(120, { min: 35, max: 700, step: 1 }, { description: 'Number of spheres/directions to test for membrane placement. Original value is 350.' }),
    stepSize: param_definition_1.ParamDefinition.Numeric(1, { min: 0.25, max: 4, step: 0.25 }, { description: 'Thickness of membrane slices that will be tested' }),
    minThickness: param_definition_1.ParamDefinition.Numeric(20, { min: 10, max: 30, step: 1 }, { description: 'Minimum membrane thickness used during refinement' }),
    maxThickness: param_definition_1.ParamDefinition.Numeric(40, { min: 30, max: 50, step: 1 }, { description: 'Maximum membrane thickness used during refinement' }),
    asaCutoff: param_definition_1.ParamDefinition.Numeric(40, { min: 10, max: 100, step: 1 }, { description: 'Absolute ASA cutoff above which residues will be considered' })
};
/**
 * Implements:
 * Membrane positioning for high- and low-resolution protein structures through a binary classification approach
 * Guillaume Postic, Yassine Ghouzam, Vincent Guiraud, and Jean-Christophe Gelly
 * Protein Engineering, Design & Selection, 2015, 1–5
 * doi: 10.1093/protein/gzv063
 */
function computeANVIL(structure, props) {
    var _this = this;
    return mol_task_1.Task.create('Compute Membrane Orientation', function (runtime) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
        return tslib_1.__generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, calculate(runtime, structure, props)];
                case 1: return [2 /*return*/, _a.sent()];
            }
        });
    }); });
}
exports.computeANVIL = computeANVIL;
var l = structure_1.StructureElement.Location.create(void 0);
var centroidHelper = new centroid_helper_1.CentroidHelper();
function initialize(structure, props) {
    var _a = structure_1.StructureProperties.atom, label_atom_id = _a.label_atom_id, x = _a.x, y = _a.y, z = _a.z;
    var elementCount = structure.polymerResidueCount;
    centroidHelper.reset();
    l.structure = structure;
    var offsets = new Int32Array(elementCount);
    var exposed = new Array(elementCount);
    var accessibleSurfaceArea = structure && accessible_surface_area_1.AccessibleSurfaceAreaProvider.get(structure);
    var asa = accessibleSurfaceArea.value;
    var vec = linear_algebra_1.Vec3();
    var m = 0;
    for (var i = 0, il = structure.units.length; i < il; ++i) {
        var unit = structure.units[i];
        var elements = unit.elements;
        l.unit = unit;
        for (var j = 0, jl = elements.length; j < jl; ++j) {
            var eI = elements[j];
            l.element = eI;
            // consider only amino acids
            if (util_1.getElementMoleculeType(unit, eI) !== 5 /* Protein */) {
                continue;
            }
            // only CA is considered for downstream operations
            if (label_atom_id(l) !== 'CA') {
                continue;
            }
            // while iterating use first pass to compute centroid
            linear_algebra_1.Vec3.set(vec, x(l), y(l), z(l));
            centroidHelper.includeStep(vec);
            // keep track of offsets and exposed state to reuse
            offsets[m] = structure.serialMapping.getSerialIndex(l.unit, l.element);
            exposed[m] = shrake_rupley_1.AccessibleSurfaceArea.getValue(l, asa) > props.asaCutoff;
            m++;
        }
    }
    // omit potentially empty tail1
    offsets = offsets.slice(0, m);
    exposed = exposed.slice(0, m);
    // calculate centroid and extent
    centroidHelper.finishedIncludeStep();
    var centroid = centroidHelper.center;
    for (var k = 0, kl = offsets.length; k < kl; k++) {
        setLocation(l, structure, offsets[k]);
        linear_algebra_1.Vec3.set(vec, x(l), y(l), z(l));
        centroidHelper.radiusStep(vec);
    }
    var extent = 1.2 * Math.sqrt(centroidHelper.radiusSq);
    return tslib_1.__assign(tslib_1.__assign({}, props), { structure: structure, offsets: offsets, exposed: exposed, centroid: centroid, extent: extent });
}
function calculate(runtime, structure, params) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var label_comp_id, ctx, initialHphobHphil, initialMembrane, alternativeMembrane, membrane;
        return tslib_1.__generator(this, function (_a) {
            label_comp_id = structure_1.StructureProperties.atom.label_comp_id;
            ctx = initialize(structure, params);
            initialHphobHphil = HphobHphil.filtered(ctx, label_comp_id);
            initialMembrane = findMembrane(ctx, generateSpherePoints(ctx, ctx.numberOfSpherePoints), initialHphobHphil, label_comp_id);
            alternativeMembrane = findMembrane(ctx, findProximateAxes(ctx, initialMembrane), initialHphobHphil, label_comp_id);
            membrane = initialMembrane.qmax > alternativeMembrane.qmax ? initialMembrane : alternativeMembrane;
            return [2 /*return*/, {
                    planePoint1: membrane.planePoint1,
                    planePoint2: membrane.planePoint2,
                    normalVector: membrane.normalVector,
                    radius: ctx.extent,
                    centroid: ctx.centroid
                }];
        });
    });
}
exports.calculate = calculate;
var MembraneCandidate;
(function (MembraneCandidate) {
    function initial(c1, c2, stats) {
        return {
            planePoint1: c1,
            planePoint2: c2,
            stats: stats
        };
    }
    MembraneCandidate.initial = initial;
    function scored(spherePoint, c1, c2, stats, qmax, centroid) {
        var diam_vect = linear_algebra_1.Vec3();
        linear_algebra_1.Vec3.sub(diam_vect, centroid, spherePoint);
        return {
            planePoint1: c1,
            planePoint2: c2,
            stats: stats,
            normalVector: diam_vect,
            spherePoint: spherePoint,
            qmax: qmax
        };
    }
    MembraneCandidate.scored = scored;
})(MembraneCandidate || (MembraneCandidate = {}));
function findMembrane(ctx, spherePoints, initialStats, label_comp_id) {
    var centroid = ctx.centroid, stepSize = ctx.stepSize, minThickness = ctx.minThickness, maxThickness = ctx.maxThickness;
    // best performing membrane
    var membrane;
    // score of the best performing membrane
    var qmax = 0;
    // construct slices of thickness 1.0 along the axis connecting the centroid and the spherePoint
    var diam = linear_algebra_1.Vec3();
    for (var i = 0, il = spherePoints.length; i < il; i++) {
        var spherePoint = spherePoints[i];
        linear_algebra_1.Vec3.sub(diam, centroid, spherePoint);
        linear_algebra_1.Vec3.scale(diam, diam, 2);
        var diamNorm = linear_algebra_1.Vec3.magnitude(diam);
        var qvartemp = [];
        var _loop_1 = function (i_1, il_1) {
            var c1 = linear_algebra_1.Vec3();
            var c2 = linear_algebra_1.Vec3();
            linear_algebra_1.Vec3.scaleAndAdd(c1, spherePoint, diam, i_1 / diamNorm);
            linear_algebra_1.Vec3.scaleAndAdd(c2, spherePoint, diam, (i_1 + stepSize) / diamNorm);
            // evaluate how well this membrane slice embeddeds the peculiar residues
            var stats = HphobHphil.filtered(ctx, label_comp_id, function (testPoint) { return isInMembranePlane(testPoint, diam, c1, c2); });
            qvartemp.push(MembraneCandidate.initial(c1, c2, stats));
        };
        for (var i_1 = 0, il_1 = diamNorm - stepSize; i_1 < il_1; i_1 += stepSize) {
            _loop_1(i_1, il_1);
        }
        var jmax = (minThickness / stepSize) - 1;
        for (var width = 0, widthl = maxThickness; width < widthl;) {
            var imax = qvartemp.length - 1 - jmax;
            for (var i_2 = 0, il_2 = imax; i_2 < il_2; i_2++) {
                var c1 = qvartemp[i_2].planePoint1;
                var c2 = qvartemp[i_2 + jmax].planePoint2;
                var hphob = 0;
                var hphil = 0;
                var total = 0;
                for (var j = 0; j < jmax; j++) {
                    var ij = qvartemp[i_2 + j];
                    if (j === 0 || j === jmax - 1) {
                        hphob += 0.5 * ij.stats.hphob;
                        hphil += 0.5 * ij.stats.hphil;
                    }
                    else {
                        hphob += ij.stats.hphob;
                        hphil += ij.stats.hphil;
                    }
                    total += ij.stats.total;
                }
                var stats = HphobHphil.of(hphob, hphil, total);
                if (hphob !== 0) {
                    var qvaltest = qValue(stats, initialStats);
                    if (qvaltest > qmax) {
                        qmax = qvaltest;
                        membrane = MembraneCandidate.scored(spherePoint, c1, c2, HphobHphil.of(hphob, hphil, total), qmax, centroid);
                    }
                }
            }
            jmax++;
            width = (jmax + 1) * stepSize;
        }
    }
    return membrane;
}
function qValue(currentStats, initialStats) {
    if (initialStats.hphob < 1) {
        initialStats.hphob = 0.1;
    }
    if (initialStats.hphil < 1) {
        initialStats.hphil += 1;
    }
    var part_tot = currentStats.hphob + currentStats.hphil;
    return (currentStats.hphob * (initialStats.hphil - currentStats.hphil) - currentStats.hphil * (initialStats.hphob - currentStats.hphob)) /
        Math.sqrt(part_tot * initialStats.hphob * initialStats.hphil * (initialStats.hphob + initialStats.hphil - part_tot));
}
function isInMembranePlane(testPoint, normalVector, planePoint1, planePoint2) {
    var d1 = -linear_algebra_1.Vec3.dot(normalVector, planePoint1);
    var d2 = -linear_algebra_1.Vec3.dot(normalVector, planePoint2);
    var d = -linear_algebra_1.Vec3.dot(normalVector, testPoint);
    return d > Math.min(d1, d2) && d < Math.max(d1, d2);
}
exports.isInMembranePlane = isInMembranePlane;
// generates a defined number of points on a sphere with radius = extent around the specified centroid
function generateSpherePoints(ctx, numberOfSpherePoints) {
    var centroid = ctx.centroid, extent = ctx.extent;
    var points = [];
    var oldPhi = 0, h, theta, phi;
    for (var k = 1, kl = numberOfSpherePoints + 1; k < kl; k++) {
        h = -1 + 2 * (k - 1) / (numberOfSpherePoints - 1);
        theta = Math.acos(h);
        phi = (k === 1 || k === numberOfSpherePoints) ? 0 : (oldPhi + 3.6 / Math.sqrt(numberOfSpherePoints * (1 - h * h))) % (2 * Math.PI);
        var point = linear_algebra_1.Vec3.create(extent * Math.sin(phi) * Math.sin(theta) + centroid[0], extent * Math.cos(theta) + centroid[1], extent * Math.cos(phi) * Math.sin(theta) + centroid[2]);
        points[k - 1] = point;
        oldPhi = phi;
    }
    return points;
}
// generates sphere points close to that of the initial membrane
function findProximateAxes(ctx, membrane) {
    var numberOfSpherePoints = ctx.numberOfSpherePoints, extent = ctx.extent;
    var points = generateSpherePoints(ctx, 30000);
    var j = 4;
    var sphere_pts2 = [];
    while (sphere_pts2.length < numberOfSpherePoints) {
        var d = 2 * extent / numberOfSpherePoints + j;
        var dsq = d * d;
        sphere_pts2 = [];
        for (var i = 0, il = points.length; i < il; i++) {
            if (linear_algebra_1.Vec3.squaredDistance(points[i], membrane.spherePoint) < dsq) {
                sphere_pts2.push(points[i]);
            }
        }
        j += 0.2;
    }
    return sphere_pts2;
}
var HphobHphil;
(function (HphobHphil) {
    function of(hphob, hphil, total) {
        return {
            hphob: hphob,
            hphil: hphil,
            total: !!total ? total : hphob + hphil
        };
    }
    HphobHphil.of = of;
    var testPoint = linear_algebra_1.Vec3();
    function filtered(ctx, label_comp_id, filter) {
        var offsets = ctx.offsets, exposed = ctx.exposed, structure = ctx.structure;
        var _a = structure_1.StructureProperties.atom, x = _a.x, y = _a.y, z = _a.z;
        var hphob = 0;
        var hphil = 0;
        for (var k = 0, kl = offsets.length; k < kl; k++) {
            // ignore buried residues
            if (!exposed[k]) {
                continue;
            }
            setLocation(l, structure, offsets[k]);
            linear_algebra_1.Vec3.set(testPoint, x(l), y(l), z(l));
            // testPoints have to be in putative membrane layer
            if (filter && !filter(testPoint)) {
                continue;
            }
            if (isHydrophobic(label_comp_id(l))) {
                hphob++;
            }
            else {
                hphil++;
            }
        }
        return of(hphob, hphil);
    }
    HphobHphil.filtered = filtered;
})(HphobHphil || (HphobHphil = {}));
// ANVIL-specific (not general) definition of membrane-favoring amino acids
var HYDROPHOBIC_AMINO_ACIDS = new Set(['ALA', 'CYS', 'GLY', 'HIS', 'ILE', 'LEU', 'MET', 'PHE', 'SER', 'THR', 'VAL']);
function isHydrophobic(label_comp_id) {
    return HYDROPHOBIC_AMINO_ACIDS.has(label_comp_id);
}
exports.isHydrophobic = isHydrophobic;
function setLocation(l, structure, serialIndex) {
    l.structure = structure;
    l.unit = structure.units[structure.serialMapping.unitIndices[serialIndex]];
    l.element = structure.serialMapping.elementIndices[serialIndex];
    return l;
}
//# sourceMappingURL=algorithm.js.map