"use strict";
/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.CoordsChange = exports.MoveLoci = void 0;
var tslib_1 = require("tslib");
var param_definition_1 = require("../../mol-util/param-definition");
var behavior_1 = require("../../mol-plugin/behavior/behavior");
var loci_1 = require("../../mol-model/loci");
var input_observer_1 = require("../../mol-util/input/input-observer");
var binding_1 = require("../../mol-util/binding");
var objects_1 = require("../../mol-plugin-state/objects");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var mol_util_1 = require("../../mol-util");
var mol_state_1 = require("../../mol-state");
var structure_1 = require("../../mol-model/structure");
var int_1 = require("../../mol-data/int");
var object_1 = require("../../mol-canvas3d/controls/object");
var B = input_observer_1.ButtonsType;
var M = input_observer_1.ModifiersKeys;
var Trigger = binding_1.Binding.Trigger;
var DefaultMoveLociBindings = {
    dragMove: binding_1.Binding([
        Trigger(1 /* Primary */, M.create({ alt: true })),
    ], 'Move Loci', 'Drag using ${triggers}'),
};
var MoveLociParams = {
    bindings: param_definition_1.ParamDefinition.Value(DefaultMoveLociBindings, { isHidden: true }),
};
exports.MoveLoci = behavior_1.PluginBehavior.create({
    name: 'move-loci',
    category: 'interaction',
    display: { name: 'Move Loci on Canvas' },
    ctor: /** @class */ (function (_super) {
        tslib_1.__extends(class_1, _super);
        function class_1() {
            return _super !== null && _super.apply(this, arguments) || this;
        }
        class_1.prototype.register = function () {
            var _this = this;
            this.ctx.state.data.actions.add(PatchedModel);
            var busy = false;
            var pendingDir;
            var currentLoci = loci_1.EmptyLoci;
            var dir = linear_algebra_1.Vec3();
            var m3 = linear_algebra_1.Mat3();
            var ref = linear_algebra_1.Vec3();
            this.subscribeObservable(this.ctx.behaviors.interaction.drag, function (_a) {
                var current = _a.current, buttons = _a.buttons, modifiers = _a.modifiers, pageStart = _a.pageStart, pageEnd = _a.pageEnd;
                if (!_this.ctx.canvas3d)
                    return;
                var binding = _this.params.bindings.dragMove;
                if (!binding_1.Binding.match(binding, buttons, modifiers))
                    return;
                if (!loci_1.Loci.areEqual(currentLoci, current.loci)) {
                    currentLoci = current.loci;
                    pendingDir = undefined;
                }
                var loci = loci_1.Loci.normalize(current.loci, _this.ctx.managers.interactivity.props.granularity);
                if (loci_1.Loci.isEmpty(loci) || !structure_1.StructureElement.Loci.is(loci))
                    return;
                loci_1.Loci.getCenter(loci, ref);
                var direction = object_1.ObjectControls.panDirection(linear_algebra_1.Vec3(), pageStart, pageEnd, _this.ctx.canvas3d.camera, ref);
                if (!busy) {
                    var _b = loci.elements[0], unit_1 = _b.unit, indices_1 = _b.indices;
                    var patchedModel = _this.ctx.state.data.select(mol_state_1.StateSelection.Generators.ofTransformer(PatchedModel)).filter(function (r) { return r.obj && structure_1.Model.getRoot(r.obj.data) === structure_1.Model.getRoot(unit_1.model); })[0];
                    if (!(patchedModel === null || patchedModel === void 0 ? void 0 : patchedModel.obj))
                        return;
                    var u_1 = unit_1.remapModel(patchedModel.obj.data);
                    if (pendingDir) {
                        linear_algebra_1.Vec3.add(dir, pendingDir, direction);
                    }
                    else {
                        linear_algebra_1.Vec3.copy(dir, direction);
                    }
                    // console.log('move', Vec3.magnitude(dir));
                    linear_algebra_1.Vec3.transformMat3(dir, dir, linear_algebra_1.Mat3.directionTransform(m3, u_1.conformation.operator.inverse));
                    var b = _this.ctx.state.data.build();
                    b.to(patchedModel.transform.ref).update(function (old) {
                        old.coords.length = 0;
                        int_1.OrderedSet.forEach(indices_1, function (v) {
                            var index = u_1.elements[v];
                            var c = u_1.conformation.invariantPosition(index, linear_algebra_1.Vec3());
                            old.coords.push({ index: index, position: linear_algebra_1.Vec3.add(c, c, dir) });
                        });
                    });
                    busy = true;
                    pendingDir = undefined;
                    b.commit({ canUndo: true, doNotLogTiming: true }).then(function () {
                        busy = false;
                    });
                }
                else if (pendingDir === undefined) {
                    // console.log('pendingPos');
                    pendingDir = direction;
                }
                else {
                    // console.log('wait');
                    linear_algebra_1.Vec3.add(pendingDir, pendingDir, direction);
                }
            });
        };
        class_1.prototype.unregister = function () {
            this.ctx.state.data.actions.remove(PatchedModel);
        };
        return class_1;
    }(behavior_1.PluginBehavior.Handler)),
    params: function () { return MoveLociParams; }
});
var PatchedModel = objects_1.PluginStateTransform.BuiltIn({
    name: 'patched-model',
    display: { name: 'Patched Model', description: 'Create a patched model.' },
    isDecorator: true,
    from: objects_1.PluginStateObject.Molecule.Model,
    to: objects_1.PluginStateObject.Molecule.Model,
    params: function (a) {
        return {
            coords: param_definition_1.ParamDefinition.Value([])
        };
    }
})({
    apply: function (_a) {
        var a = _a.a, params = _a.params, cache = _a.cache;
        if (a.data.parent)
            throw new Error('Only the root model can be patched.');
        structure_1.Model.CoordinatesHistory.set(a.data, new CoordinatesHistory());
        var _b = a.data.atomicConformation, x = _b.x, y = _b.y, z = _b.z;
        var coords = new BufferedCoordinates(x, y, z);
        cache.bufferedCoords = coords;
        if (params.coords.length === 0)
            return new objects_1.PluginStateObject.Molecule.Model(a.data);
        return new objects_1.PluginStateObject.Molecule.Model(getPatchedModel(a.data, coords.swap(), params.coords));
    },
    update: function (_a) {
        var a = _a.a, b = _a.b, oldParams = _a.oldParams, newParams = _a.newParams, cache = _a.cache;
        if (a.data !== structure_1.Model.getRoot(b.data))
            return mol_state_1.StateTransformer.UpdateResult.Recreate;
        if (mol_util_1.deepEqual(oldParams, newParams))
            return mol_state_1.StateTransformer.UpdateResult.Unchanged;
        var coords = cache.bufferedCoords;
        b.data = getPatchedModel(b.data, coords.swap(), newParams.coords);
        return mol_state_1.StateTransformer.UpdateResult.Updated;
    }
});
var CoordsChangeProp = '__CoordsChange__';
var CoordsChange = {
    set: function (model, coordsChange) {
        model._staticPropertyData[CoordsChangeProp] = coordsChange;
    },
    get: function (model) {
        return model._staticPropertyData[CoordsChangeProp];
    }
};
exports.CoordsChange = CoordsChange;
var CoordinatesHistory = /** @class */ (function () {
    function CoordinatesHistory() {
    }
    CoordinatesHistory.prototype.areEqual = function (elements, kind, model) {
        var coordsChange = CoordsChange.get(structure_1.Model.getRoot(model));
        if (coordsChange) {
            return !int_1.SortedArray.areIntersecting(elements, coordsChange);
        }
        return false;
    };
    return CoordinatesHistory;
}());
function getPatchedModel(model, coords, patch) {
    var x = coords.x, y = coords.y, z = coords.z;
    var root = structure_1.Model.getRoot(model);
    var coordsChange = [];
    for (var _i = 0, patch_1 = patch; _i < patch_1.length; _i++) {
        var _a = patch_1[_i], index = _a.index, position = _a.position;
        x[index] = position[0];
        y[index] = position[1];
        z[index] = position[2];
        coordsChange.push(index);
    }
    CoordsChange.set(root, int_1.SortedArray.ofUnsortedArray(coordsChange));
    var atomicConformation = tslib_1.__assign(tslib_1.__assign({}, root.atomicConformation), { id: mol_util_1.UUID.create22(), x: x, y: y, z: z });
    return tslib_1.__assign(tslib_1.__assign({}, root), { id: mol_util_1.UUID.create22(), parent: root, atomicConformation: atomicConformation });
}
var BufferedCoordinates = /** @class */ (function () {
    function BufferedCoordinates(x, y, z) {
        this.x = new BufferedNumberArray(x, Float32Array);
        this.y = new BufferedNumberArray(y, Float32Array);
        this.z = new BufferedNumberArray(z, Float32Array);
    }
    BufferedCoordinates.prototype.swap = function () {
        return {
            x: this.x.swap(),
            y: this.y.swap(),
            z: this.z.swap()
        };
    };
    return BufferedCoordinates;
}());
var BufferedNumberArray = /** @class */ (function () {
    function BufferedNumberArray(orig, ctor) {
        this.orig = orig;
        this.ctor = ctor;
    }
    BufferedNumberArray.prototype.swap = function () {
        if (!this.prev)
            this.prev = new this.ctor(this.orig);
        if (!this.next)
            this.next = new this.ctor(this.prev);
        var ret = this.next;
        ret.set(this.prev);
        this.next = this.prev;
        this.prev = ret;
        return ret;
    };
    return BufferedNumberArray;
}());
//# sourceMappingURL=move-loci.js.map