"use strict";
/**
 * Copyright (c) 2018-2019 Mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Types = void 0;
var tslib_1 = require("tslib");
var type_1 = tslib_1.__importDefault(require("../type"));
var Core = tslib_1.__importStar(require("./core"));
var symbol_1 = require("../symbol");
var helpers_1 = require("../helpers");
var Types;
(function (Types) {
    Types.ElementSymbol = type_1.default.Value('Structure', 'ElementSymbol');
    Types.AtomName = type_1.default.Value('Structure', 'AtomName');
    Types.BondFlag = type_1.default.OneOf('Structure', 'BondFlag', type_1.default.Str, ['covalent', 'metallic', 'ion', 'hydrogen', 'sulfide', 'computed', 'aromatic']);
    Types.BondFlags = Core.Types.Flags(Types.BondFlag, 'BondFlags');
    Types.SecondaryStructureFlag = type_1.default.OneOf('Structure', 'SecondaryStructureFlag', type_1.default.Str, ['alpha', 'beta', '3-10', 'pi', 'sheet', 'strand', 'helix', 'turn', 'none']);
    Types.SecondaryStructureFlags = Core.Types.Flags(Types.SecondaryStructureFlag, 'SecondaryStructureFlag');
    Types.RingFingerprint = type_1.default.Value('Structure', 'RingFingerprint');
    Types.EntityType = type_1.default.OneOf('Structure', 'EntityType', type_1.default.Str, ['polymer', 'non-polymer', 'water', 'branched']);
    Types.EntitySubtype = type_1.default.OneOf('Structure', 'EntitySubtype', type_1.default.Str, ['other', 'polypeptide(D)', 'polypeptide(L)', 'polydeoxyribonucleotide', 'polyribonucleotide', 'polydeoxyribonucleotide/polyribonucleotide hybrid', 'cyclic-pseudo-peptide', 'peptide nucleic acid', 'oligosaccharide']);
    Types.ObjectPrimitive = type_1.default.OneOf('Structure', 'ObjectPrimitive', type_1.default.Str, ['atomistic', 'sphere', 'gaussian', 'other']);
    Types.ResidueId = type_1.default.Value('Structure', 'ResidueId');
    Types.ElementSet = type_1.default.Value('Structure', 'ElementSet');
    Types.ElementSelection = type_1.default.Value('Structure', 'ElementSelection');
    Types.ElementReference = type_1.default.Value('Structure', 'ElementReference');
    Types.ElementSelectionQuery = Core.Types.Fn(Types.ElementSelection, 'ElementSelectionQuery');
})(Types = exports.Types || (exports.Types = {}));
var type = {
    '@header': 'Types',
    elementSymbol: helpers_1.symbol(symbol_1.Arguments.Dictionary({ 0: symbol_1.Argument(type_1.default.Str) }), Types.ElementSymbol, 'Create element symbol representation from a string value.'),
    atomName: helpers_1.symbol(symbol_1.Arguments.Dictionary({ 0: symbol_1.Argument(type_1.default.AnyValue) }), Types.AtomName, 'Convert a value to an atom name.'),
    entityType: helpers_1.symbol(symbol_1.Arguments.Dictionary({ 0: symbol_1.Argument(Types.EntityType) }), Types.EntityType, "Create normalized representation of entity type: " + type_1.default.oneOfValues(Types.EntityType).join(', ') + "."),
    bondFlags: helpers_1.symbol(symbol_1.Arguments.List(Types.BondFlag), Types.BondFlags, "Create bond flags representation from a list of strings. Allowed flags: " + type_1.default.oneOfValues(Types.BondFlag).join(', ') + "."),
    ringFingerprint: helpers_1.symbol(symbol_1.Arguments.List(Types.ElementSymbol, { nonEmpty: true }), Types.RingFingerprint, 'Create ring fingerprint from the supplied atom element list.'),
    secondaryStructureFlags: helpers_1.symbol(symbol_1.Arguments.List(Types.SecondaryStructureFlag), Types.SecondaryStructureFlags, "Create secondary structure flags representation from a list of strings. Allowed flags: " + type_1.default.oneOfValues(Types.SecondaryStructureFlag).join(', ') + "."),
    authResidueId: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(type_1.default.Str, { description: 'auth_asym_id' }),
        1: symbol_1.Argument(type_1.default.Num, { description: 'auth_seq_id' }),
        2: symbol_1.Argument(type_1.default.Str, { description: 'pdbx_PDB_ins_code', isOptional: true })
    }), Types.ResidueId, "Residue identifier based on \"auth_\" annotation."),
    labelResidueId: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(type_1.default.Str, { description: 'label_entity_id' }),
        1: symbol_1.Argument(type_1.default.Str, { description: 'label_asym_id' }),
        2: symbol_1.Argument(type_1.default.Num, { description: 'label_seq_id' }),
        3: symbol_1.Argument(type_1.default.Str, { description: 'pdbx_PDB_ins_code', isOptional: true })
    }), Types.ResidueId, "Residue identifier based on mmCIF's \"label_\" annotation.")
};
var slot = {
    '@header': 'Iteration Slots',
    element: helpers_1.symbol(symbol_1.Arguments.None, Types.ElementReference, 'A reference to the current element.'),
    elementSetReduce: helpers_1.symbol(symbol_1.Arguments.None, type_1.default.Variable('a', type_1.default.AnyValue, true), 'Current value of the element set reducer.')
};
var generator = {
    '@header': 'Generators',
    all: helpers_1.symbol(symbol_1.Arguments.None, Types.ElementSelectionQuery, 'The entire structure.'),
    atomGroups: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        'entity-test': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: true, description: 'Test for the 1st atom of every entity' }),
        'chain-test': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: true, description: 'Test for the 1st atom of every chain' }),
        'residue-test': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: true, description: 'Test for the 1st atom every residue' }),
        'atom-test': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: true }),
        'group-by': symbol_1.Argument(type_1.default.Any, { isOptional: true, defaultValue: "atom-key", description: 'Group atoms to sets based on this property. Default: each atom has its own set' }),
    }), Types.ElementSelectionQuery, 'Return all atoms for which the tests are satisfied, grouped into sets.'),
    bondedAtomicPairs: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: 'true for covalent bonds', description: 'Test each bond with this predicate. Each bond is visited twice with swapped atom order.' }),
    }), Types.ElementSelectionQuery, 'Return all pairs of atoms for which the test is satisfied.'),
    rings: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        'fingerprint': symbol_1.Argument(Types.RingFingerprint, { isOptional: true }),
        'only-aromatic': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: false }),
    }), Types.ElementSelectionQuery, 'Return all rings or those with the specified fingerprint and/or only aromatic rings.'),
    queryInSelection: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        query: symbol_1.Argument(Types.ElementSelectionQuery),
        'in-complement': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: false })
    }), Types.ElementSelectionQuery, 'Executes query only on atoms that are in the source selection.'),
    empty: helpers_1.symbol(symbol_1.Arguments.None, Types.ElementSelectionQuery, 'Nada.'),
};
var modifier = {
    '@header': 'Selection Modifications',
    queryEach: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        query: symbol_1.Argument(Types.ElementSelectionQuery)
    }), Types.ElementSelectionQuery, 'Query every atom set in the input selection separately.'),
    intersectBy: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        by: symbol_1.Argument(Types.ElementSelectionQuery)
    }), Types.ElementSelectionQuery, 'Intersect each atom set from the first sequence from atoms in the second one.'),
    exceptBy: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        by: symbol_1.Argument(Types.ElementSelectionQuery)
    }), Types.ElementSelectionQuery, "Remove all atoms from 'selection' that occur in 'by'."),
    unionBy: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        by: symbol_1.Argument(Types.ElementSelectionQuery)
    }), Types.ElementSelectionQuery, 'For each atom set A in the orginal sequence, combine all atoms sets in the target selection that intersect with A.'),
    union: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery)
    }), Types.ElementSelectionQuery, 'Collects all atom sets in the sequence into a single atom set.'),
    cluster: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        'min-distance': symbol_1.Argument(type_1.default.Num, { isOptional: true, defaultValue: 0 }),
        'max-distance': symbol_1.Argument(type_1.default.Num),
        'min-size': symbol_1.Argument(type_1.default.Num, { description: 'Minimal number of sets to merge, must be at least 2', isOptional: true, defaultValue: 2 }),
        'max-size': symbol_1.Argument(type_1.default.Num, { description: 'Maximal number of sets to merge, if not set, no limit', isOptional: true }),
    }), Types.ElementSelectionQuery, 'Combines atom sets that have mutual distance in the interval [min-radius, max-radius]. Minimum/maximum size determines how many atom sets can be combined.'),
    includeSurroundings: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        radius: symbol_1.Argument(type_1.default.Num),
        'atom-radius': symbol_1.Argument(type_1.default.Num, { isOptional: true, defaultValue: 0, description: 'Value added to each atom before the distance check, for example VDW radius. Using this argument is computationally demanding.' }),
        'as-whole-residues': symbol_1.Argument(type_1.default.Bool, { isOptional: true })
    }), Types.ElementSelectionQuery, 'For each atom set in the selection, include all surrouding atoms/residues that are within the specified radius.'),
    includeConnected: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        'bond-test': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: 'true for covalent bonds' }),
        'layer-count': symbol_1.Argument(type_1.default.Num, { isOptional: true, defaultValue: 1, description: 'Number of bonded layers to include.' }),
        'as-whole-residues': symbol_1.Argument(type_1.default.Bool, { isOptional: true })
    }), Types.ElementSelectionQuery, 'Pick all atom sets that are connected to the target.'),
    wholeResidues: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
    }), Types.ElementSelectionQuery, 'Expand the selection to whole residues.'),
    expandProperty: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        property: symbol_1.Argument(type_1.default.AnyValue)
    }), Types.ElementSelectionQuery, 'To each atom set in the selection, add all atoms that have the same property value that was already present in the set.')
};
var filter = {
    '@header': 'Selection Filters',
    pick: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        test: symbol_1.Argument(type_1.default.Bool)
    }), Types.ElementSelectionQuery, 'Pick all atom sets that satisfy the test.'),
    first: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery)
    }), Types.ElementSelectionQuery, 'Take the 1st atom set in the sequence.'),
    withSameAtomProperties: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        source: symbol_1.Argument(Types.ElementSelectionQuery),
        property: symbol_1.Argument(type_1.default.Any)
    }), Types.ElementSelectionQuery, 'Pick all atom sets for which the set of given atom properties is a subset of the source properties.'),
    intersectedBy: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        by: symbol_1.Argument(Types.ElementSelectionQuery)
    }), Types.ElementSelectionQuery, 'Pick all atom sets that have non-zero intersection with the target.'),
    within: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        target: symbol_1.Argument(Types.ElementSelectionQuery),
        'min-radius': symbol_1.Argument(type_1.default.Num, { isOptional: true, defaultValue: 0 }),
        'max-radius': symbol_1.Argument(type_1.default.Num),
        'atom-radius': symbol_1.Argument(type_1.default.Num, { isOptional: true, defaultValue: 0, description: 'Value added to each atom before the distance check, for example VDW radius. Using this argument is computationally demanding.' }),
        invert: symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: false, description: 'If true, pick only atom sets that are further than the specified radius.' }),
    }), Types.ElementSelectionQuery, 'Pick all atom sets from selection that have any atom within the radius of any atom from target.'),
    isConnectedTo: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery),
        target: symbol_1.Argument(Types.ElementSelectionQuery),
        'bond-test': symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: 'true for covalent bonds' }),
        disjunct: symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: true, description: 'If true, there must exist a bond to an atom that lies outside the given atom set to pass test.' }),
        invert: symbol_1.Argument(type_1.default.Bool, { isOptional: true, defaultValue: false, description: 'If true, return atom sets that are not connected.' })
    }), Types.ElementSelectionQuery, 'Pick all atom sets that are connected to the target.'),
};
var combinator = {
    '@header': 'Selection Combinators',
    intersect: helpers_1.symbol(symbol_1.Arguments.List(Types.ElementSelectionQuery), Types.ElementSelectionQuery, 'Return all unique atom sets that appear in all of the source selections.'),
    merge: helpers_1.symbol(symbol_1.Arguments.List(Types.ElementSelectionQuery), Types.ElementSelectionQuery, 'Merges multiple selections into a single one. Only unique atom sets are kept.'),
    distanceCluster: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        matrix: symbol_1.Argument(Core.Types.List(Core.Types.List(type_1.default.Num)), { description: 'Distance matrix, represented as list of rows (num[][])). Lower triangle is min distance, upper triangle is max distance.' }),
        selections: symbol_1.Argument(Core.Types.List(Types.ElementSelectionQuery), { description: 'A list of held selections.' })
    }), Types.ElementSelectionQuery, 'Pick combinations of atom sets from the source sequences that are mutually within distances specified by a matrix.')
};
var atomSet = {
    '@header': 'Atom Sets',
    atomCount: helpers_1.symbol(symbol_1.Arguments.None, type_1.default.Num),
    countQuery: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Types.ElementSelectionQuery)
    }), type_1.default.Num, 'Counts the number of occurences of a specific query inside the current atom set.'),
    reduce: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        initial: symbol_1.Argument(type_1.default.Variable('a', type_1.default.AnyValue, true), { description: 'Initial value assigned to slot.atom-set-reduce. Current atom is set to the 1st atom of the current set for this.' }),
        value: symbol_1.Argument(type_1.default.Variable('a', type_1.default.AnyValue, true), { description: 'Expression executed for each atom in the set' })
    }), type_1.default.Variable('a', type_1.default.AnyValue, true), 'Execute the value expression for each atom in the current atom set and return the result. Works the same way as Array.reduce in JavaScript (``result = value(value(...value(initial)))``)'),
    propertySet: helpers_1.symbol(symbol_1.Arguments.Dictionary({
        0: symbol_1.Argument(Core.Types.ConstrainedVar),
    }), Core.Types.Set(Core.Types.ConstrainedVar), 'Returns a set with all values of the given property in the current atom set.'),
};
var atomProperty = {
    '@header': 'Atom Properties',
    core: {
        '@header': 'Core Properties',
        elementSymbol: atomProp(Types.ElementSymbol),
        vdw: atomProp(type_1.default.Num, 'Van der Waals radius'),
        mass: atomProp(type_1.default.Num, 'Atomic weight'),
        atomicNumber: atomProp(type_1.default.Num, 'Atomic number'),
        x: atomProp(type_1.default.Num, 'Cartesian X coordinate'),
        y: atomProp(type_1.default.Num, 'Cartesian Y coordinate'),
        z: atomProp(type_1.default.Num, 'Cartesian Z coordinate'),
        atomKey: atomProp(type_1.default.AnyValue, 'Unique value for each atom. Main use case is grouping of atoms.'),
        bondCount: helpers_1.symbol(symbol_1.Arguments.Dictionary({
            0: symbol_1.Argument(Types.ElementReference, { isOptional: true, defaultValue: 'slot.current-atom' }),
            flags: symbol_1.Argument(Types.BondFlags, { isOptional: true, defaultValue: 'covalent' }),
        }), type_1.default.Num, 'Number of bonds (by default only covalent bonds are counted).'),
        sourceIndex: atomProp(type_1.default.Num, 'Index of the atom/element in the input file.'),
        operatorName: atomProp(type_1.default.Str, 'Name of the symmetry operator applied to this element.'),
        modelIndex: atomProp(type_1.default.Num, 'Index of the model in the input file.'),
        modelLabel: atomProp(type_1.default.Str, 'Label/header of the model in the input file.')
    },
    topology: {
        connectedComponentKey: atomProp(type_1.default.AnyValue, 'Unique value for each connected component.')
    },
    macromolecular: {
        '@header': 'Macromolecular Properties (derived from the mmCIF format)',
        authResidueId: atomProp(Types.ResidueId, "type.auth-residue-id symbol executed on current atom's residue"),
        labelResidueId: atomProp(Types.ResidueId, "type.label-residue-id symbol executed on current atom's residue"),
        residueKey: atomProp(type_1.default.AnyValue, 'Unique value for each tuple ``(label_entity_id,auth_asym_id, auth_seq_id, pdbx_PDB_ins_code)``, main use case is grouping of atoms'),
        chainKey: atomProp(type_1.default.AnyValue, 'Unique value for each tuple ``(label_entity_id, auth_asym_id)``, main use case is grouping of atoms'),
        entityKey: atomProp(type_1.default.AnyValue, 'Unique value for each tuple ``label_entity_id``, main use case is grouping of atoms'),
        isHet: atomProp(type_1.default.Bool, 'Equivalent to atom_site.group_PDB !== ATOM'),
        id: atomProp(type_1.default.Num, '_atom_site.id'),
        label_atom_id: atomProp(Types.AtomName),
        label_alt_id: atomProp(type_1.default.Str),
        label_comp_id: atomProp(type_1.default.Str),
        label_asym_id: atomProp(type_1.default.Str),
        label_entity_id: atomProp(type_1.default.Str),
        label_seq_id: atomProp(type_1.default.Num),
        auth_atom_id: atomProp(Types.AtomName),
        auth_comp_id: atomProp(type_1.default.Str),
        auth_asym_id: atomProp(type_1.default.Str),
        auth_seq_id: atomProp(type_1.default.Num),
        pdbx_PDB_ins_code: atomProp(type_1.default.Str),
        pdbx_formal_charge: atomProp(type_1.default.Num),
        occupancy: atomProp(type_1.default.Num),
        B_iso_or_equiv: atomProp(type_1.default.Num),
        entityType: atomProp(Types.EntityType, 'Type of the entity as defined in mmCIF (polymer, non-polymer, branched, water)'),
        entitySubtype: atomProp(Types.EntitySubtype, 'Subtype of the entity as defined in mmCIF _entity_poly.type and _pdbx_entity_branch.type (other, polypeptide(D), polypeptide(L), polydeoxyribonucleotide, polyribonucleotide, polydeoxyribonucleotide/polyribonucleotide hybrid, cyclic-pseudo-peptide, peptide nucleic acid, oligosaccharide)'),
        entityPrdId: atomProp(type_1.default.Str, "The PRD ID of the entity."),
        entityDescription: atomProp(Core.Types.List(type_1.default.Str)),
        objectPrimitive: atomProp(Types.ObjectPrimitive, 'Type of the primitive object used to model this segment as defined in mmCIF/IHM (atomistic, sphere, gaussian, other)'),
        secondaryStructureKey: atomProp(type_1.default.AnyValue, 'Unique value for each secondary structure element.'),
        secondaryStructureFlags: atomProp(Types.SecondaryStructureFlags),
        isModified: atomProp(type_1.default.Bool, 'True if the atom belongs to modification of a standard residue.'),
        modifiedParentName: atomProp(type_1.default.Str, "'3-letter' code of the modifed parent residue."),
        isNonStandard: atomProp(type_1.default.Bool, 'True if this is a non-standard residue.'),
        chemCompType: atomProp(type_1.default.Str, "Type of the chemical component as defined in mmCIF."),
    }
};
var bondProperty = {
    '@header': 'Bond Properties',
    flags: bondProp(Types.BondFlags),
    order: bondProp(type_1.default.Num),
    length: bondProp(type_1.default.Num),
    atomA: bondProp(Types.ElementReference),
    atomB: bondProp(Types.ElementReference)
};
function atomProp(type, description) {
    return helpers_1.symbol(symbol_1.Arguments.Dictionary({ 0: symbol_1.Argument(Types.ElementReference, { isOptional: true, defaultValue: 'slot.current-atom' }) }), type, description);
}
function bondProp(type, description) {
    return helpers_1.symbol(symbol_1.Arguments.None, type, description);
}
exports.default = {
    '@header': 'Structure Queries',
    type: type,
    slot: slot,
    generator: generator,
    modifier: modifier,
    filter: filter,
    combinator: combinator,
    atomSet: atomSet,
    atomProperty: atomProperty,
    bondProperty: bondProperty
};
//# sourceMappingURL=structure-query.js.map